/******************************************************************************
*
*  FILENAME :	efac.c
*
*  DESCRIPTION :
*
*	Source code for the EFAC client, efac.
*
*  PUBLIC FUNCTIONS :
*
*	int main(int argc, char *argv[], char **envp)
*
*  NOTES :
*
*	Written by Paul Chesson (1994)
*	E-Mail: chesson@mundil.cs.mu.OZ.AU
*
*/


/******************************************************************************
*  INCLUDE FILES
******************************************************************************/

#include <errno.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "bugbyte.h"
#include "file.h"
#include "version.h"


/******************************************************************************
*  PRIVATE DECLARATIONS
******************************************************************************/

/* ----- data declarations ------------------------------------------------- */

#define USERNAME_MAX	8		/* maximum length of user names	*/
#define SERVER_PATH	"/.efacs/efacs"	/* path of server from home dir	*/
#define DEFAULT_TMP	"/tmp"		/* default temporary directory	*/

/* ----- function prototypes ----------------------------------------------- */

static char	*server_name(char *efacs_owner, char *file);

/* ----- macros ------------------------------------------------------------ */

#define STR_EQ3(A, B, C) (STR_EQ((A), (B)) || STR_EQ((A), (C)))
#define IS_A (STR_EQ3(argv[1], "a", "-a"))
#define IS_C (STR_EQ3(argv[1], "c", "-c"))
#define IS_D (STR_EQ3(argv[1], "d", "-d"))
#define IS_I (STR_EQ3(argv[1], "i", "-i"))
#define IS_R (STR_EQ3(argv[1], "r", "-r"))
#define IS_W (STR_EQ3(argv[1], "w", "-w"))
#define IS_X (STR_EQ3(argv[1], "x", "-x"))

#define IS_ACW (IS_A || IS_C || IS_W)
#define IS_DIR (IS_D || IS_I || IS_R)


/******************************************************************************
*  PUBLIC FUNCTIONS
******************************************************************************/

/******************************************************************************
*
*  NAME :	int main(int argc, char *argv[], char **envp)
*
*  DESCRIPTION :
*
*	Front-end client for Extend File Access Control (EFAC).
*	Calls appropriate EFAC server and manipulates the output.
*
*  INPUTS :
*
*	PARAMETERS:	argc		number of command line arguments
*			argv[]		array of command line arguments
*			envp		enviroment strings
*		
*  OUTPUTS :
*
*	RETURN:		OK		normal program termination
*			otherwise	abnormal termination
*
*  NOTES :
*
*	Environment required to access $PATH for exec.
*
*/

int main(int argc, char *argv[], char **envp)
{
    char	efacs_owner[USERNAME_MAX+1];

    envp = envp;			/* keeps lint happy */

    /*
     *      Extract out options, leaving main command
     */

    if (argc >= 3 && STR_EQ(argv[1], "-u"))
    {
	int i;

	if (strlen(argv[2]) > USERNAME_MAX)
	{
	    fprintf(stderr, "%s: User name too long.\n", argv[0]);
	    exit(-1);
	}

	strcpy(efacs_owner, argv[2]);
	for (i = 1; i < argc-2; i++)
	    argv[i] = argv[i+2];
	argv[argc-2] = NULL;
	argc -= 2;
    }


    /*
     *	    Command Line Processing.
     *	    Exit with usage message if command line syntax incorrect.
     */

    if (argc < 3 || (argc > 3 && IS_ACW))
    {
	fprintf(stderr, "%s %s, %s\n", argv[0], EFAC_VERSION, COPYRIGHT);
	fprintf(stderr, "Usage: %s [-u user] [-]{acw} file\n", argv[0]);
	fprintf(stderr, "       %s [-u user] [-]{dir} file ...\n", argv[0]);
	fprintf(stderr, "       %s [-u user] [-]x file [args]\n", argv[0]);
	fprintf(stderr, "       %s [-u user] command [args] file\n", argv[0]);
	exit(-1);
    }


    /*
     *	    Handle EFAC operation.
     *	    Make call to server and handle output where necessary.
     */

    /* _____ append, copy, and write _____ */

    if (IS_ACW)
    {
	execl(server_name(efacs_owner, argv[2]), "efacs", argv[1], argv[2],
	    NULL);
	FATAL("execl");
    }

    /* _____ delete, info, and read _____ */

    else if (IS_DIR)
    {
	int i;

	for (i = 2; i < argc; i++)
	{
            pid_t pid;

	    if ((pid = fork()) == -1)
		FATAL(argv[0]);

	    if (pid == 0)
	    {
		if (IS_I)
		{
		    if (i > 2)
			printf("\n");
		    if (argc > 3)
			printf("%s:\n", argv[i]);
		}
	        execl(server_name(efacs_owner, argv[i]), "efacs", argv[1],
		    argv[i], NULL);
		FATAL("execl");
	    }

	    wait(NULL);

	}  /* for each file */
    }  /* delete, info, read */

    /* _____ execute command _____ */

    else
    {
	char		*file;		/* file name requested by user	*/
        const char	**aux_args;	/* arg to construct efacs call	*/
        char		tmp_file[FILENAME_MAX+1]; /* temp copy of file	*/
        pid_t		pid;

	MALLOC_TYPE(aux_args, const char *, argc);
        sprintf(tmp_file, "%s/efac%u",
		getenv("TMPDIR") ? getenv("TMPDIR") : DEFAULT_TMP,
		(unsigned int)getpid());

	/* _____ execute file _____ */

	if (IS_X)
	{
	    int i;

	    file = argv[2];
	    aux_args[0] = tmp_file;
	    for (i = 1; i < argc-2; i++)
		aux_args[i] = argv[i+2];
	    aux_args[argc-2] = NULL;
	}

	/* _____ use file in command _____ */

	else
	{
	    int i;

	    file = argv[argc-1];
	    for (i = 0; i < argc-2; i++)
		aux_args[i] = argv[i+1];
	    aux_args[argc-2] = tmp_file;
	    aux_args[argc-1] = NULL;
	}

	/* _____ read file into temp file _____ */

	if ((pid = fork()) == -1)
	    FATAL(argv[0]);

	if (pid == 0)
	{
	    reopen_file(tmp_file, "w", stdout);
	    chmod(tmp_file, S_IRWXU);
	    execl(server_name(efacs_owner, file), "efacs", "r", file, NULL);
	    FATAL(argv[0]);
	}

	wait(NULL);

	/* _____ perform execution on temp file _____ */

	if ((pid = fork()) == -1)
	    FATAL(argv[0]);

	if (pid == 0)
	{
	    execvp(aux_args[0], (char **)aux_args);
	    FATAL(argv[0]);
	}

	wait(NULL);

	/* _____ copy temp file back if it has changed _____ */

	if (diff_file(file, tmp_file) == TRUE)
	{
	    if ((pid = fork()) == -1)
		FATAL(argv[0]);

	    if (pid == 0)
	    {
		reopen_file(tmp_file, "r", stdin);
		execl(server_name(efacs_owner, file), "efacs", "w", file,
		      NULL);
		FATAL(argv[0]);
	    }

	    wait(NULL);

	    if (diff_file(file, tmp_file) == FALSE)
		delete_file(tmp_file);
	    else
		fprintf(stderr, "Modified file stored as %s.\n", tmp_file);
	}
	else
	{
	    delete_file(tmp_file);
	}
	
    }  /* execute command */

    return OK;

}  /* main() */


/******************************************************************************
*  PRIVATE FUNCTIONS
******************************************************************************/

/******************************************************************************
*
*  NAME :	char *server_name(char *efacs_owner, char *file)
*
*  DESCRIPTION :
*
*	Determines which server should be called for a given file.
*
*  INPUTS :
*
*	PARAMETERS:	efacs_owner	explicit name of user requested
*					  (empty string if none)
*			filename	name of file requested by user
*		
*  OUTPUTS :
*
*	RETURN:		Full path name of server to use.
*
*  NOTES :
*
*	Returned name is statically stored, so should be copied if
*	required for later use.
*
*/

static char *server_name(char *efacs_owner, char *filename)
{
    static char	server_file[FILENAME_MAX+1];	/* returned efacs path	*/
    struct stat	statbuf;
    uid_t	uid;

    /*
     *	    Determine user ID of efacs owner to be called from
     *	    the prioritised list (1) - (5).
     */

    /* _____ (1)  Use specified server _____ */

    if (STR_DIFF(efacs_owner, ""))
    {
	struct passwd *pp;

	if ((pp = getpwnam(efacs_owner)) == NULL)
	{
	    fprintf(stderr, "%s: Unknown user %s.\n", filename, efacs_owner);
	    exit(-1);
	}
	uid = pp->pw_uid;
    }
    
    /* _____ (2)  Use owner of requested file _____*/

    else if (stat(filename, &statbuf) == 0)
    {
	uid = getpwuid(statbuf.st_uid)->pw_uid;
    }

    /*
     *	    (3)  Use owner of requested file's directory's ACL.
     *      (4)  Use owner of requested file's ACL.
     *	    (5)  Use requesting user.
     */

    else
    {
        char	dir_acl[FILENAME_MAX+1];	/* (3) */
	char	file_acl[FILENAME_MAX+1];	/* (4) */

	strcpy(dir_acl, get_dir(filename));
	strcat(dir_acl, ".acl");
	strcpy(file_acl, filename);
	strcat(file_acl, ".acl");

	if ( (stat(dir_acl, &statbuf) == 0)
	    || (stat(file_acl, &statbuf) == 0) )
	{
	    uid = getpwuid(statbuf.st_uid)->pw_uid;
	}
	else
	{
	    uid = getuid();			/* (5) */
	}
    }

    /*
     *	    Derive full path name for server, and check that it has
     *	    set UID (prevents trojan horses being called).
     */

    strcpy(server_file, getpwuid(uid)->pw_dir);
    strcat(server_file, SERVER_PATH);
    if (stat(server_file, &statbuf) < 0)
	FATAL(server_file);

    if ((statbuf.st_mode & S_ISUID) == 0)
    {
	fprintf(stderr, "%s: set user ID permission off\n", server_file);
	exit(EACCES);
    }

    return server_file;
}

